



set CMAKE_GENERATOR=Ninja
IF %ERRORLEVEL% NEQ 0 exit /B 1
cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_INSTALL_PREFIX=%LIBRARY_PREFIX% -DPython3_EXECUTABLE=%PYTHON% -DPYTHON_EXECUTABLE=%PYTHON% -DTBB_TEST=ON -DTBB_DIR=TRUE -S . -B test_build
IF %ERRORLEVEL% NEQ 0 exit /B 1
cmake --build test_build --target test_parallel_for test_tbb_header conformance_version test_malloc_compliance test_flow_graph test_arena_constraints --parallel
IF %ERRORLEVEL% NEQ 0 exit /B 1
cd test_build
IF %ERRORLEVEL% NEQ 0 exit /B 1
ctest -R "^(test_parallel_for|test_tbb_header|conformance_version|test_malloc_compliance|test_flow_graph|test_arena_constraints)$" --output-on-failure
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
